from dataclasses import dataclass
import logging

import littletable as db

from .constants import *
from .datastructures import UniqueRecordByWhere

logger = logging.getLogger(__name__)

# Device profiles
@dataclass
class DeviceProfile:
    name: str
    address: int
    device_id: int
    capabilities: list[int]
    typ: int
    interface: int = ADAFRUIT

    def is_mux(self) -> bool:
        return True if self.typ == MUX else False

    def is_meter(self) -> bool:
        return True if self.typ == METER else False

    def is_detector(self) -> bool:
        return True if self.typ == DETECTOR else False

    def has_capability(self, cap: int) -> bool:
        return True if cap in self.capabilities else False

profiles = db.Table('profiles')
profiles.insert(DeviceProfile('PCA9546A', 0x70, PCA9546A, [ FOUR_CHANNEL ], MUX))
# FIXME need discovery discrimination before this will work
#profiles.insert(DeviceProfile('TCA9548A', 0x70, TCA9548A, [ EIGHT_CHANNEL ], MUX))
profiles.insert(DeviceProfile('BMP390', 0x77, BMP390, [ PRESSURE, TEMPERATURE, ALTITUDE ], METER))
profiles.insert(DeviceProfile('SHT41', 0x44, SHT41, [ TEMPERATURE, RELATIVE_HUMIDITY ], METER))
profiles.insert(DeviceProfile('VEML7700', 0x10, VEML7700, [ AMBIENT_LIGHT, LUX ], METER))
profiles.insert(DeviceProfile('SCD41', 0x62, SCD41, [ CO2, RELATIVE_HUMIDITY, TEMPERATURE ], METER))
profiles.insert(DeviceProfile('TSL2591', 0x29, TSL2591,
                              [ LUX, FULL_SPECTRUM, VISIBLE, INFRARED, AMBIENT_LIGHT ], METER))

profiles.create_index('name')
profiles.create_index('address')
profiles.create_index('device_id')
profile_selector = UniqueRecordByWhere(profiles)



import abc
from collections.abc import Iterator
import logging
from typing import Literal

logger = logging.getLogger(__name__)

import board
from busio import I2C
import adafruit_tca9548a

from . import constants
from . import devices
from .tools.mixins import NodeMixin

class ChannelProxy(NodeMixin):
    def __init__(self, index: int, channel):
        super().__init__()
        self._index = index
        self._channel = channel

    @property
    def channel_id(self) -> int:
        return self._index

    def __getattr__(self, attr):
        return getattr(self._channel, attr)

class MuxInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'address') and
                callable(subclass.address) and
                hasattr(subclass, 'channels') and
                callable(subclass.channels) and
                hasattr(subclass, '__len__') and
                callable(subclass.__len__) or
                NotImplemented)

    @abc.abstractmethod
    def channels(self) -> Iterator[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError

class PCA9546A(NodeMixin, devices.Device, MuxInterface):
    def __init__(self, bus: I2C, name: str, device_id: int, capabilities: list[int],
                 address: int | str = 112):
        super().__init__(bus, name, device_id, capabilities, address)
        self._mux = adafruit_tca9548a.PCA9546A(bus, address)
        self._channels = [None] * 4

        for i in range(len(self._mux)):
            self._channels[i] = ChannelProxy(i, self._mux[i])

    def __len__(self) -> Literal[4]:
        return len(self._mux)

    @property
    def real_device(self):
        return self._mux

    def channels(self) -> Iterator[int]:
        for c in self._channels:
            yield c

class MuxFactory:
    def __init__(self):
        self._ctors = {}

    def register_mux(self, device_id, ctor):
        self._ctors[device_id] = ctor

    def get_mux(self, bus: I2C, name: str, device_id: int, capabilities: list[int],
                address: int | str) -> MuxInterface:
        ctor = self._ctors.get(device_id)
        if not ctor:
            raise ValueError(device_id)

        return ctor(bus, name, device_id, capabilities, address)

mux_factory = MuxFactory()
mux_factory.register_mux(constants.PCA9546A, PCA9546A)



import abc
import logging
from typing import Any
import urllib.parse
import urllib.request
import uuid

logger = logging.getLogger(__name__)

class NodeMixin():
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._uuid = uuid.uuid4()

    @property
    def uuid(self) -> uuid.UUID:
        return str(self._uuid)

class GetterMixin(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, '_handler') and
                callable(subclass._handler) and
                hasattr(subclass, 'location') and
                callable(subclass.location) or
                NotImplemented)

    def url_get(self, params: dict) -> None:
        endpoint = self.location + urllib.parse.urlencode(params)

        logger.debug('calling api endpoint %s', endpoint)
        contents = urllib.request.urlopen(endpoint)
        self._handler(contents)

    @abc.abstractmethod
    def location(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _handler(self, state, contents):
        raise NotImplementedError

class SchedulableInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'schedule') and
                callable(subclass.schedule) and
                hasattr(subclass, 'unschedule') and
                callable(subclass.unschedule) or
                NotImplemented)

    @abc.abstractmethod
    def schedule(self, immediate: bool) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def unschedule(self) -> None:
        raise NotImplementedError

class RunnableInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'run') and
                callable(subclass.run) and
                hasattr(subclass, 'stop') and
                callable(subclass.stop) or
                NotImplemented)

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError



NONE = 0x00

# Addresses
VIRTUAL_ADDR = 0xFF

# Interfaces
ADAFRUIT = 0x0001
DFROBOT  = 0x0002

# Device Types
VIRTUAL       = 0x80000000

BUS           = 0x0001
MUX           = 0x0002
CHANNEL       = 0x0003
DEVICE        = 0x0004
METER         = 0x0005
DETECTOR      = 0x0006

# Capabilities
FOUR_CHANNEL            = 0x0001
EIGHT_CHANNEL           = 0x0002
PRESSURE                = 0x0004
TEMPERATURE             = 0x0008
ALTITUDE                = 0x0010
RELATIVE_HUMIDITY       = 0x0020
AMBIENT_LIGHT           = 0x0040
LUX                     = 0x0080
VISIBLE                 = 0x0100
INFRARED                = 0x0200
FULL_SPECTRUM           = 0x0400
CO2                     = 0x0800
PRESSURE_MSL            = 0x1000

# Units
CELSIUS_UNITS                = 'Celsius (C)'
HECTOPASCAL_UNITS            = 'Hectopascal (hPa)'
METER_UNITS                  = 'Meter (M)'
PERC_RELATIVE_HUMIDITY_UNITS = 'Percent Relative Humidity (% rH)'
AMBIENT_LIGHT_UNITS          = 'Ambient Light Data'
LUX_UNITS                    = 'Lux (Lx)'
PPM_UNITS                    = 'Parts per Million (PPM)'

# Breakout boards / DeviceIDs
VIRTUAL_DEVICE = 0xFFFF
PCA9546A       = 0x0001
PCA9548A       = 0x0002
TCA9548A       = PCA9548A
BMP390         = 0x0003
SHT41          = 0x0004
VEML7700       = 0x0005
SCD41          = 0x0006
TSL2591        = 0x0007



import abc
import json
import logging
from typing import Any

from isodate import parse_duration
import littletable as db

from ..constants import (
        VIRTUAL_ADDR,
        VIRTUAL_DEVICE,
        PRESSURE_MSL,
        TEMPERATURE,
        RELATIVE_HUMIDITY,
)
from ..datastructures import (UniqueRecordFieldByWhere,
                              capabilities_selector,
)
from ..meters import MeterInterface
from ..tools.mixins import (
        GetterMixin,
        SchedulableInterface,
        NodeMixin,
)
from .virtual import Virtual

logger = logging.getLogger(__name__)

open_meteo_data = f"""\
id,capability,open_meteo_capability
{TEMPERATURE},temperature,temperature_2m
{RELATIVE_HUMIDITY},relative_humidity,relative_humidity_2m"""

open_meteo = db.Table('open_meteo')
open_meteo.csv_import(open_meteo_data, transforms={"id": int})

fields = UniqueRecordFieldByWhere(open_meteo)

class _OpenMeteoInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, '_handler') and
                callable(subclass._handler) or
                NotImplemented)

    @abc.abstractmethod
    def _handler(self) -> None:
        raise NotImplementedError

class _OpenMeteoCurrentGetterImpl(GetterMixin, SchedulableInterface):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super(_OpenMeteoCurrentGetterImpl, cls).__new__(cls)
        return cls._instance
    
    def __init__(self, capabilities: list[str], interval: str, params: dict[str, int | str],
                 scheduler):
        self._location = 'https://api.open-meteo.com/v1/forecast?'
        self._handlers = {}

        self._interval = parse_duration(interval).seconds
        self._params = params
        self._capabilities = capabilities
        self._scheduler = scheduler
        self._job = None

        param = [ c if not fields('open_meteo_capability', capability=c).found else \
                fields('open_meteo_capability', capability=c).field for c in capabilities ]
        self._params['current'] = ','.join(param)

    def schedule(self, immediate: bool) -> None:
        if self._job is not None:
            return

        if immediate:
            self.url_get(self._params)

        self._job = self._scheduler.add_job(self.url_get, 'interval', seconds=self._interval,
                                            kwargs = { 'params': self._params })

    def unschedule(self) -> None:
        if self._job is None:
            return

        self._job.remove()
        self._job = None

    def set_handler(self, capability: str, handler: callable) -> None:
        if capability not in self._capabilities:
            logger.warning('open meteo impl: requested unset cap: %s', capability)
            return

        self._handlers[capability] = handler

    @property
    def location(self) -> str:
        return self._location

    def _handler(self, contents) -> None:
        if contents.status != 200:
            logger.warning('open meteo impl: API failed GET - %s', contents.status)
            return

        data = contents.read()
        obj = json.loads(data)

        current_values = obj['current']
        current_units = obj['current_units']

        for c in self._handlers:
            func = self._handlers[c]
            om_cap = fields('open_meteo_capability', capability=c)
            v = current_values[c] if not om_cap.found else current_values[om_cap.field]
            u = current_units[c] if not om_cap.found  else current_units[om_cap.field]
            func(c, v, u)

class _OpenMeteoCurrent(NodeMixin, Virtual, _OpenMeteoInterface):
    def __init__(self, name: str, capability: str, interval: str,
                 params: dict[str, int | str], scheduler):
        super().__init__(name, capability)
        self._measure = 0.0
        self._units = None

    @property
    def measure(self) -> float:
        return self._measure

    @property
    def measurement(self) -> int:
        return self._measurement

    @property
    def units(self) -> str:
        return self._units

    def _handler(self, capability: str, value: float | int, units: str) -> None:
        logger.debug('open_meteo_handler called with %s %s %s',
                     capability, value, units)

        if capability != self._capability:
            msg = 'open meteo meter - called with mismatch capability: got %s != wanted %s'
            logger.warning(msg, capability, self._capability)
            return

        self._measure = value
        self._units = units

class OpenMeteoCurrentBuilder:
    def __init__(self, name: str, capabilities: list[str]):
        self._name = name
        self._caps = capabilities

    def __call__(self, interval: str, params: dict[str, int | str], scheduler,
                 **_ignored) -> list[_OpenMeteoCurrent | _OpenMeteoCurrentGetterImpl]:
        # the builder knows what this is, put your smarts here when building other layered
        # virtual devices

        getter_impl = _OpenMeteoCurrentGetterImpl(self._caps, interval, params, scheduler)

        devs = []
        for cap in self._caps:
            obj = _OpenMeteoCurrent(self._name, cap, interval, params, scheduler)
            getter_impl.set_handler(cap, obj._handler)
            devs.append(obj)

        getter_impl.schedule(True)

        return devs



from ..constants import (
        VIRTUAL_ADDR,
        VIRTUAL_DEVICE,
)
from ..datastructures import (
        capabilities_selector,
)
from ..meters import MeterInterface

class Virtual(MeterInterface):
    def __init__(self, name: str, capability: str):
        self._name = name
        field = capabilities_selector('id', capability=capability)
        if field.found is False:
            raise ValueError('unsupported capability {}'.format(capability))
        self._measurement = field.field
        self._capability = capability

    @property
    def address(self) -> int:
        return VIRTUAL_ADDR

    @property
    def device_id(self) -> int:
        return VIRTUAL_DEVICE

    @property
    def name(self) -> str:
        return self._name

    @property
    def channel_id(self) -> [ int | None ]:
        return None



import abc
from collections.abc import Iterator
import logging
from typing import Any

import adafruit_bmp3xx
import adafruit_scd4x
import adafruit_sht4x
import adafruit_tsl2591
import adafruit_veml7700
from busio import I2C

from . import constants
from . import profiles
from .tools.mixins import (
        NodeMixin,
        RunnableInterface,
)

logger = logging.getLogger(__name__)

class DeviceCapabilityError(Exception):
    """Raised when Device cannot read requested capability."""

class DeviceInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'address') and
                callable(subclass.address) and
                hasattr(subclass, 'device_id') and
                callable(subclass.device_id) and
                hasattr(subclass, 'name') and
                callable(subclass.name) and
                hasattr(subclass, 'channel_id') and
                callable(subclass.channel_id) and
                hasattr(subclass, 'real_device') and
                callable(subclass.real_device) and
                hasattr(subclass, 'capabilities') and
                callable(subclass.capabilities) and
                hasattr(subclass, 'capabilities_gen') and
                callable(subclass.capabilities_gen) and
                hasattr(subclass, 'read_capability') and
                callable(subclass.read_capability) and
                hasattr(subclass, 'capability_units') and
                callable(subclass.capability_units) or
                NotImplemented)

    @abc.abstractmethod
    def address(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def device_id(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def channel_id(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def real_device(self):
        raise NotImplementedError

    @abc.abstractmethod
    def capabilities(self) -> list[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def capabilities_gen(self) -> Iterator[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def read_capability(self, capability: int) -> [ int | float ]:
        raise NotImplementedError

    @abc.abstractmethod
    def capability_units(self, capability: int) -> str:
        raise NotImplementedError

class Device(DeviceInterface):
    def __init__(self, bus: I2C | None, name: str, device_id: int, capabilities: list[int],
                 address: int, env: dict[str, Any] | None = None):
        self._bus = bus
        self._name = name
        self._device_id = device_id
        self._caps = capabilities
        self._dev = None
        self._property_map = dict()
        self._capability_units = dict()

        self._address = address
        self._has_channel = False
        self._channel_id = 0
        if self._bus is not None and hasattr(self._bus, 'channel_switch'):
            n = int.from_bytes(self._bus.channel_switch, 'little')
            r = 0
            while n > 1:
                n = n >> 1
                r = r + 1
            self._channel_id = r
            self._has_channel = True

    @property
    def address(self) -> int:
        return self._address

    @property
    def device_id(self) -> int:
        return self._device_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def capabilities(self) -> list[int]:
        return self._caps

    def capabilities_gen(self) -> Iterator[int]:
        for cap in self._caps:
            yield cap

    def read_capability(self, capability: int) -> [ int | float ]:
        try:
            prop = self._property_map[capability]
        except KeyError:
            raise DeviceCapabilityError('capability ({}) is unsupported by device ({})'.format(
                capability, self._device_id))

        if self._dev is None:
            msg = 'device initialized without real device ({})'.format(self._device_id)
            raise DeviceCapabilityError(msg)

        return getattr(self._dev, prop)

    def capability_units(self, capability: int) -> str:
        try:
            units = self._capability_units[capability]
            return units
        except KeyError:
            msg = 'capability ({}) has no associated units for device ({})'.format(capability,
                                                                                   self._device_id)
            raise DeviceCapabilityError(msg)

    @property
    def channel_id(self) -> [ int | None ]:
        if self._has_channel is True:
            return self._channel_id
        return None

class Bmp390(NodeMixin, Device):
    def __init__(self, bus: I2C, name: str, device_id: int,
                 address: int = 119, env: dict[str, Any] | None = None):
        super().__init__(bus, name, device_id,
                         [ constants.TEMPERATURE, constants.PRESSURE, constants.ALTITUDE ],
                         address)
        self._dev = adafruit_bmp3xx.BMP3XX_I2C(bus, address)
        self._property_map[constants.TEMPERATURE] = 'temperature'
        self._capability_units[constants.TEMPERATURE] = constants.CELSIUS_UNITS
        self._property_map[constants.PRESSURE] = 'pressure'
        self._capability_units[constants.PRESSURE] = constants.HECTOPASCAL_UNITS
        self._property_map[constants.ALTITUDE] = 'altitude'
        self._capability_units[constants.ALTITUDE] = constants.METER_UNITS

    @property
    def real_device(self):
        return self._dev

class Sht41(NodeMixin, Device):
    def __init__(self, bus: I2C, name: str, device_id: int,
                 address: int = 68, env: dict[str, Any] | None = None):
        super().__init__(bus, name, device_id,
                         [ constants.RELATIVE_HUMIDITY, constants.TEMPERATURE ],
                         address)
        self._dev = adafruit_sht4x.SHT4x(bus, address)
        self._property_map[constants.RELATIVE_HUMIDITY] = 'relative_humidity'
        self._capability_units[constants.RELATIVE_HUMIDITY] = constants.PERC_RELATIVE_HUMIDITY_UNITS
        self._property_map[constants.TEMPERATURE] = 'temperature'
        self._capability_units[constants.TEMPERATURE] = constants.CELSIUS_UNITS

    @property
    def real_device(self):
        return self._dev

class Veml7700(NodeMixin, Device):
    def __init__(self, bus: I2C, name: str, device_id: int,
                 address: int = 16, env: dict[str, Any] | None = None):
        super().__init__(bus, name, device_id, [ constants.LUX, constants.AMBIENT_LIGHT ], address)
        self._dev = adafruit_veml7700.VEML7700(bus, address)
        self._property_map[constants.LUX] = 'lux'
        self._capability_units[constants.LUX] = constants.LUX_UNITS
        self._property_map[constants.AMBIENT_LIGHT] = 'light'
        self._capability_units[constants.AMBIENT_LIGHT] = constants.AMBIENT_LIGHT_UNITS

    @property
    def real_device(self):
        return self._dev

class Scd41(NodeMixin, Device, RunnableInterface):
    def __init__(self, bus: I2C, name: str, device_id: int,
                 address: int = 98, env: dict[str, Any] | None = None):
        super().__init__(bus, name, device_id,
                         [ constants.TEMPERATURE, constants.RELATIVE_HUMIDITY, constants.CO2 ],
                         address)
        self._dev = adafruit_scd4x.SCD4X(bus, address)
        self._property_map[constants.TEMPERATURE] = 'temperature'
        self._capability_units[constants.TEMPERATURE] = constants.CELSIUS_UNITS
        self._property_map[constants.RELATIVE_HUMIDITY] = 'relative_humidity'
        self._capability_units[constants.RELATIVE_HUMIDITY] = constants.PERC_RELATIVE_HUMIDITY_UNITS
        self._property_map[constants.CO2] = 'CO2'
        self._capability_units[constants.CO2] = constants.PPM_UNITS

        self._dev.reinit()
        self._env = env

    @property
    def real_device(self):
        return self._dev

    def run(self):
        if 'indoors' in self._env:
            enable = not self._env['indoors']
            self._dev.self_calibration_enabled = enable

        self._dev.start_periodic_measurement()

    def stop(self):
        self._dev.stop_periodic_measurement()

class Tsl2591(NodeMixin, Device):
    def __init__(self, bus: I2C, name: str, device_id: int,
                 address: int = 41, env: dict[str, Any] | None = None):
        super().__init__(bus, name, device_id,
                         [ constants.LUX, constants.INFRARED, constants.VISIBLE,
                           constants.FULL_SPECTRUM ],
                         address)
        self._dev = adafruit_tsl2591.TSL2591(bus, address)
        self._property_map[constants.LUX] = 'lux'
        self._capability_units[constants.LUX] = constants.LUX_UNITS
        self._property_map[constants.INFRARED] = 'infrared'
        self._capability_units[constants.INFRARED] = constants.AMBIENT_LIGHT_UNITS
        self._property_map[constants.VISIBLE] = 'visible'
        self._capability_units[constants.INFRARED] = constants.AMBIENT_LIGHT_UNITS
        self._property_map[constants.FULL_SPECTRUM] = 'full_spectrum'
        self._capability_units[constants.FULL_SPECTRUM] = constants.AMBIENT_LIGHT_UNITS

    @property
    def real_device(self):
        return self._dev

class DeviceFactory:
    def __init__(self):
        self._ctors = {}

    def register_device(self, device_id, ctor):
        self._ctors[device_id] = ctor

    def get_device(self, bus: I2C, name: str, device_id: int,
                   address: int, env: dict[str, Any] | None = None) -> DeviceInterface:
        ctor = self._ctors.get(device_id)
        if not ctor:
            raise ValueError(device_id)

        return ctor(bus, name, device_id, address, env)

device_factory = DeviceFactory()
device_factory.register_device(constants.BMP390, Bmp390)
device_factory.register_device(constants.SHT41, Sht41)
device_factory.register_device(constants.VEML7700, Veml7700)
device_factory.register_device(constants.SCD41, Scd41)
device_factory.register_device(constants.TSL2591, Tsl2591)



from collections import namedtuple
import logging
import typing

import littletable as db

from . import constants

logger = logging.getLogger(__name__)

interfaces_data = f"""\
id,interface
{constants.ADAFRUIT},adafruit
{constants.DFROBOT},dfrobot"""

interfaces = db.Table('interfaces')
interfaces.create_index('id', unique=True)
interfaces.create_index('interface', unique=True)
interfaces.csv_import(interfaces_data, transforms={'id': int})

device_type_data = f"""\
type,device
{constants.BUS},bus
{constants.MUX},mux
{constants.CHANNEL},channel
{constants.DEVICE},device
{constants.METER},meter
{constants.DETECTOR},detector"""

device_types = db.Table('device_types')
device_types.create_index('type', unique=True)
device_types.csv_import(device_type_data, transforms={'type': int})

capabilities_data = f"""\
id,capability
{constants.FOUR_CHANNEL},four_channel
{constants.EIGHT_CHANNEL},eight_channel
{constants.PRESSURE},pressure
{constants.TEMPERATURE},temperature
{constants.ALTITUDE},altitude
{constants.RELATIVE_HUMIDITY},relative_humidity
{constants.AMBIENT_LIGHT},ambient_light
{constants.LUX},lux
{constants.VISIBLE},visible
{constants.INFRARED},infrared
{constants.FULL_SPECTRUM},full_spectrum
{constants.CO2},co2
{constants.PRESSURE_MSL},pressure_msl"""

capabilities = db.Table('capabilities')
capabilities.create_index('id', unique=True)
capabilities.create_index('capability', unique=True)
capabilities.csv_import(capabilities_data, transforms={'id': int})

device_ids_data = f"""\
id,device_name
{constants.VIRTUAL_DEVICE},VIRTUAL
{constants.PCA9546A},PCA9546A
{constants.PCA9548A},PCA9548A
{constants.BMP390},BMP390
{constants.SHT41},SHT41
{constants.VEML7700},VEML7700
{constants.SCD41},SCD41
{constants.TSL2591},TSL2591"""

device_ids = db.Table('device_ids')
device_ids.create_index('id', unique=True)
device_ids.create_index('device_name', unique=True)
device_ids.csv_import(device_ids_data, transforms={'id': int})

links = db.Table('links')
links.create_index('node', unique=True)
links.create_index('parent')

nodes = db.Table('nodes')
nodes.create_index('uuid', unique=True)
nodes.create_index('kind')

multiplexer_attributes = db.Table('multiplexer_attributes')
multiplexer_attributes.create_index('uuid', unique=True)
channel_attributes = db.Table('channel_attributes')
channel_attributes.create_index('uuid', unique=True)
device_attributes = db.Table('device_attributes')
device_attributes.create_index('uuid', unique=True)
device_attributes.create_index('name')
device_attributes.create_index('device_id')
device_attributes.create_index('channel_id')
device_attributes.create_index('address')
meter_attributes = db.Table('meter_attributes')
meter_attributes.create_index('uuid', unique=True)
meter_attributes.create_index('measurement')
virtual_attributes = db.Table('virtual_attributes')
virtual_attributes.create_index('uuid', unique=True)
virtual_attributes.create_index('name')
virtual_attributes.create_index('measurement')

# Joins
def join_devices():
    devices = device_attributes.outer_join(join_type=db.Table.FULL_OUTER_JOIN,
                                           other=nodes,
                                           attrlist=[
                                               (nodes, 'uuid'),
                                               (nodes, 'kind'),
                                               (nodes, 'obj'),
                                               (nodes, 'is_virtual'),
                                               (device_attributes, 'uuid'),
                                               (device_attributes, '_address', 'address'),
                                               (device_attributes, '_has_channel', 'has_channel'),
                                               (device_attributes, '_channel_id', 'channel_id'),
                                               (device_attributes, '_name', 'name'),
                                               (device_attributes, '_caps', 'capabilities'),
                                               (device_attributes, '_device_id', 'device_id'),
                                           ],
                                           uuid='uuid')('devices')
    return devices.where(kind=constants.DEVICE)

def join_meters():
    meters = meter_attributes.outer_join(join_type=db.Table.FULL_OUTER_JOIN,
                                         other=nodes,
                                         attrlist=[
                                             (nodes, 'uuid'),
                                             (nodes, 'kind'),
                                             (nodes, 'obj'),
                                             (nodes, 'is_virtual'),
                                             (meter_attributes, 'uuid'),
                                             (meter_attributes, '_measurement', 'measurement'),
                                         ],
                                         uuid='uuid')('meters')
    return meters.where(kind=constants.METER)

def join_virtuals():
    virtuals = virtual_attributes.outer_join(join_type=db.Table.FULL_OUTER_JOIN,
                                             other=nodes,
                                             attrlist=[
                                                 (nodes, 'uuid'),
                                                 (nodes, 'kind'),
                                                 (nodes, 'obj'),
                                                 (nodes, 'is_virtual'),
                                                 (virtual_attributes, 'uuid'),
                                                 (virtual_attributes, 'name'),
                                                 (virtual_attributes, 'measurement'),
                                             ],
                                             uuid='uuid')('virtuals')
    return virtuals.where(is_virtual=True)

def join_devices_meters():
    devices = device_attributes.outer_join(join_type=db.Table.FULL_OUTER_JOIN,
                                           other=links,
                                           attrlist=[
                                               (links, 'node'),
                                               (links, 'parent'),
                                               (device_attributes, 'uuid'),
                                               (device_attributes, '_name', 'name'),
                                               (device_attributes, '_address', 'address'),
                                               (device_attributes, '_has_channel', 'has_channel'),
                                               (device_attributes, '_channel_id', 'channel_id'),
                                               (device_attributes, '_device_id', 'device_id'),
                                           ],
                                           uuid='parent')('devices')

    meters = meter_attributes.outer_join(join_type=db.Table.FULL_OUTER_JOIN,
                                         other=nodes,
                                         attrlist=[
                                             (nodes, 'uuid'),
                                             (nodes, 'obj', 'meter_obj'),
                                             (meter_attributes, '_measurement', 'measurement'),
                                         ],
                                         uuid='uuid')('meters')

    children = devices.outer_join(join_type=db.Table.FULL_OUTER_JOIN,
                                  other=meters,
                                  attrlist=[
                                      (devices,'name'),
                                      (devices,'uuid'),
                                      (meters,'uuid','child_uuid'),
                                      (meters,'meter_obj'),
                                      (meters,'measurement'),
                                  ],
                                  node='uuid')('device_meters')
    return children

class NonUniqueRecordQuery(Exception):
    """Non Unique Query"""

Field = namedtuple('Field', 'found field')
Record = namedtuple('Record', 'found record')

class UniqueRecordByWhere:
    def __init__(self, table: db.Table[db.TableContent]):
        self._table = table

    def __call__(self, where: dict[str, typing.Any] = None, **kwargs) -> Record:
        clauses = dict()
        if where is not None:
            clauses = {**where, **kwargs}
        else:
            clauses = {**kwargs}
        rec = self._table.where(**clauses)
        match len(rec):
            case 0:
                return Record(False, None)
            case 1:
                return Record(True, rec[0])
            case _:
                msg = 'Table {} - clause: {} - returned non unique record'.format(
                        self._table.table_name, clauses)
                raise NonUniqueRecordQuery(msg)

class UniqueRecordFieldByWhere:
    def __init__(self, table: db.Table[db.TableContent]):
        self._table = table

    def __call__(self, field: str, where: dict[str, typing.Any] = None, **kwargs) -> Field:
        clauses = dict()
        if where is not None:
            clauses = {**where, **kwargs}
        else:
            clauses = {**kwargs}
        rec = self._table.where(**clauses)
        match len(rec):
            case 0:
                return Field(False, None)
            case 1:
                return Field(True, getattr(rec[0], field))
            case _:
                msg = 'Table {} - clause: {} - returned non unique record'.format(
                        self._table.table_name, clauses)
                raise NonUniqueRecordQuery(msg)

interfaces_selector   = UniqueRecordFieldByWhere(interfaces)
devicetypes_selector  = UniqueRecordFieldByWhere(device_types)
capabilities_selector = UniqueRecordFieldByWhere(capabilities)
deviceids_selector    = UniqueRecordFieldByWhere(device_ids)



import logging
from collections.abc import Iterator
from typing import Any

import board
from busio import I2C

from . import constants
from . import controls
from . import datastructures
from . import devices
from . import meters
from . import profiles
from .tools.mixins import NodeMixin

logger = logging.getLogger(__name__)

class DeviceTree:
    def __init__(self, i2c: I2C, env: dict[str, Any] | None = None):
        self._i2c = i2c
        self._env = env

    def build(self) -> None:
        self._build_tree(self._i2c, None, self._env)

    def add(self, obj: NodeMixin | None, kind: int, parent: NodeMixin | None):
        datastructures.links.insert({'node': obj.uuid,
                                     'parent': parent.uuid if parent is not None else '__none__'})
        datastructures.nodes.insert({'uuid': obj.uuid, 'kind': kind, 'obj': obj,
                                     'is_virtual': bool(kind & constants.VIRTUAL)})

        match kind:
            case constants.MUX:
                datastructures.multiplexer_attributes.insert(obj)
            case constants.CHANNEL:
                datastructures.channel_attributes.insert(obj)
            case constants.DEVICE:
                datastructures.device_attributes.insert(obj)
            case constants.METER:
                datastructures.meter_attributes.insert(obj)
            case _:
                if bool(kind & (constants.VIRTUAL | constants.METER)):
                    datastructures.virtual_attributes.insert(
                            {'uuid': obj.uuid, 'name': obj.name, 'measurement': obj.measurement,
                             'capability': obj._capability})
                else:
                    raise ValueError('unsupported kind {}'.format(kind))

    def _build_tree(self, i2c, parent: NodeMixin | None, env: dict[str, Any] | None = None,
                    addr_filter: set = set()):
        try:
            if i2c.try_lock():
                devs = i2c.scan()
            i2c.unlock()
        except AttributeError:
            devs = i2c.scan()

        logger.debug('initial scan results: %s, applying filter: %s',
                     [hex(n) for n in devs],
                     [hex(n) for n in addr_filter])

        devs = [n for n in devs if n not in addr_filter]

        for addr in devs:
            logger.info('building node for address: %s', hex(addr))

            try:
                record = profiles.profile_selector(address=addr)
                if record.found is False:
                    message = 'bus scan reports unsupported address {}, continuing...'.format(
                        hex(addr))
                    logger.warning(message)
                    continue

                self._build_node(i2c, addr, record.record, parent, env)
            except Exception as e:
                message = 'bus scan raised exception, {}'.format(e)
                logger.warning(message)

    def _build_node(self, i2c, address, profile, parent: NodeMixin | None,
                    env: dict[str, Any] | None = None):
        if profile.is_mux():
            logger.info('multiplexer %s found at addr %s, setting up multiple channels',
                        profile.name, hex(address))

            mux = controls.mux_factory.get_mux(i2c, profile.name, profile.device_id,
                                               profile.capabilities, address)
            logger.info('multiplexer supported channels: %s', len(mux))

            self.add(mux, constants.MUX, None)

            for channel in mux.channels():
                self.add(channel, constants.CHANNEL, mux)

                addr_set = set()
                addr_set.add(mux.address)
                self._build_tree(channel, channel, env, addr_set)
        else:
            dev = devices.device_factory.get_device(i2c, profile.name, profile.device_id,
                                                    address, env)

            self.add(dev, constants.DEVICE, parent)
            self._build_leaves(i2c, dev)

    def _build_leaves(self, i2c, parent):
        # XXX only do this for meters, eventually need to discriminate with detectors
        for cap in parent.capabilities_gen():
            try:
                m = meters.Meter(parent, cap)
                self.add(m, constants.METER, parent)
            except ValueError as e:
                capstr = datastructures.capabilities_selector('capability', id=cap)
                logger.warning('name %s, device_id %s, capability %s - no ctor',
                               device.name, device.device_id, capstr.field)



import abc
from collections.abc import Iterator
import logging
import typing

from busio import I2C

from . import datastructures
from . import devices
from .constants import *
from .tools.mixins import NodeMixin

logger = logging.getLogger(__name__)

class MeterInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'measure') and
                callable(subclass.measure) and
                hasattr(subclass, 'measurement') and
                callable(subclass.measurement) and
                hasattr(subclass, 'units') and
                callable(subclass.units) and
                hasattr(subclass, 'address') and
                callable(subclass.address) and
                hasattr(subclass, 'device_id') and
                callable(subclass.device_id) and
                hasattr(subclass, 'name') and
                callable(subclass.name) and
                hasattr(subclass, 'channel_id') and
                callable(subclass.channel_id) or
                NotImplemented)

    @abc.abstractmethod
    def measure(self) -> float:
        raise NotImplementedError

    @abc.abstractmethod
    def measurement(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def units(self) -> str:
        raise NotImplementedError

    @property
    def address(self) -> int:
        raise NotImplementedError

    @property
    def device_id(self) -> int:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def channel_id(self) -> int:
        raise NotImplementedError

class Meter(NodeMixin, MeterInterface):
    def __init__(self, device, measurement: int):
        super().__init__()
        self._device = device
        self._measurement = measurement

    @property
    def address(self) -> int:
        return self._device._address

    @property
    def device_id(self) -> int:
        return self._device._device_id

    @property
    def name(self) -> str:
        return self._device._name

    @property
    def channel_id(self) -> int:
        return self._device._channel_id

    @property
    def measure(self) -> float:
        return self._device.read_capability(self._measurement)

    @property
    def measurement(self) -> int:
        return self._measurement

    @property
    def units(self) -> str:
        return self._device.capability_units(self._measurement)
